/*
 * This file is part of roccat-tools.
 *
 * roccat-tools is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * roccat-tools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with roccat-tools. If not, see <http://www.gnu.org/licenses/>.
 */

#include "arvo_eventhandler_channel.h"
#include "g_cclosure_roccat_marshaller.h"
#include "roccat.h"
#include "arvo.h"
#include <sys/ioctl.h>

#define ARVO_EVENTHANDLER_CHANNEL_CLASS(klass) (G_TYPE_CHECK_CLASS_CAST((klass), ARVO_EVENTHANDLER_CHANNEL_TYPE, ArvoEventhandlerChannelClass))
#define IS_ARVO_EVENTHANDLER_CHANNEL_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE((klass), ARVO_EVENTHANDLER_CHANNEL_TYPE))
#define ARVO_EVENTHANDLER_CHANNEL_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE((obj), ARVO_EVENTHANDLER_CHANNEL_TYPE, ArvoEventhandlerChannelPrivate))

typedef struct _ArvoEventhandlerChannelClass ArvoEventhandlerChannelClass;
typedef struct _ArvoEventhandlerChannelPrivate ArvoEventhandlerChannelPrivate;

struct _ArvoEventhandlerChannel {
	GObject parent;
	ArvoEventhandlerChannelPrivate *priv;
};

struct _ArvoEventhandlerChannelClass {
	GObjectClass parent_class;
};

struct _ArvoEventhandlerChannelPrivate {
	guint event_source_id;
	gint report_size;
	gboolean running;
};

enum {
	BUTTON,
	LAST_SIGNAL
};

static guint signals[LAST_SIGNAL] = { 0 };

G_DEFINE_TYPE(ArvoEventhandlerChannel, arvo_eventhandler_channel, G_TYPE_OBJECT);

static void process_chrdev_event(ArvoEventhandlerChannel *eventhandler_channel, ArvoRoccatReport const *event) {
	switch (event->button) {
	case ARVO_ROCCAT_REPORT_BUTTON_MODE_ON:
		break;
	case ARVO_ROCCAT_REPORT_BUTTON_MODE_OFF:
		break;
	case ARVO_ROCCAT_REPORT_BUTTON_1:
	case ARVO_ROCCAT_REPORT_BUTTON_2:
	case ARVO_ROCCAT_REPORT_BUTTON_3:
	case ARVO_ROCCAT_REPORT_BUTTON_4:
	case ARVO_ROCCAT_REPORT_BUTTON_5:
		g_signal_emit((gpointer)eventhandler_channel, signals[BUTTON], 0, event->action, event->button);
		break;
	default:
		g_debug("unknown event = %02x%02x%02x", event->unused, event->button, event->action);
	}
}

static gboolean io_cb(GIOChannel *channel, GIOCondition condition, gpointer user_data) {
	ArvoEventhandlerChannel *eventhandler_channel = ARVO_EVENTHANDLER_CHANNEL(user_data);
	ArvoEventhandlerChannelPrivate *priv = eventhandler_channel->priv;
	gchar *event;
	GIOStatus status;
	GError *error = NULL;
	gsize length;

	if (condition & G_IO_HUP)
		return FALSE;

	event = g_malloc(priv->report_size);

	status = g_io_channel_read_chars(channel, event, priv->report_size, &length, &error);

	if (status == G_IO_STATUS_AGAIN)
		goto ok;

	if (status == G_IO_STATUS_ERROR || status == G_IO_STATUS_EOF) {
		g_critical("error reading io_channel");
		goto error;
	}

	if (length != priv->report_size) {
		g_critical("wrong length %zd %s", length, error->message);
		goto error;
	}

	switch (priv->report_size) {
	case sizeof(ArvoRoccatReport):
		process_chrdev_event(eventhandler_channel, (ArvoRoccatReport const *)event);
		break;
	default:
		g_critical("wrong event size");
		goto error;
	}

ok:
	g_free(event);
	return TRUE; // keep on going
error:
	g_free(event);
	return FALSE;
}

ArvoEventhandlerChannel *arvo_eventhandler_channel_new(void) {
	return ARVO_EVENTHANDLER_CHANNEL(g_object_new(ARVO_EVENTHANDLER_CHANNEL_TYPE, NULL));
}

static void arvo_eventhandler_channel_init(ArvoEventhandlerChannel *channel) {
	ArvoEventhandlerChannelPrivate *priv = ARVO_EVENTHANDLER_CHANNEL_GET_PRIVATE(channel);
	channel->priv = priv;
	priv->running = FALSE;
}

gboolean arvo_eventhandler_channel_start(ArvoEventhandlerChannel *eventhandler_channel, gchar const *path, GError **error) {
	ArvoEventhandlerChannelPrivate *priv = eventhandler_channel->priv;
	GIOChannel *channel;
	GIOStatus status;
	int fd;
	
	channel = g_io_channel_new_file(path, "r", error);
	if (!channel)
		return FALSE;

	status = g_io_channel_set_encoding(channel, NULL, error);
	if (status != G_IO_STATUS_NORMAL) {
		g_io_channel_unref(channel);
		return FALSE;
	}

	fd = g_io_channel_unix_get_fd(channel);
	if (ioctl(fd, ROCCATIOCGREPSIZE, &priv->report_size) == -1)
		priv->report_size = sizeof(ArvoRoccatReport);

	if (priv->report_size != sizeof(ArvoRoccatReport)) {
		g_set_error(error, G_FILE_ERROR, G_FILE_ERROR_FAILED, "ioctl reported unsupported event size %i", priv->report_size);
		g_io_channel_unref(channel);
		return FALSE;
	}

	priv->event_source_id = g_io_add_watch(channel, G_IO_IN | G_IO_HUP, (GIOFunc)io_cb, eventhandler_channel);
	/* g_io_add_watch increases reference count
	 * this passes ownership of the channel to the watch */
	g_io_channel_unref(channel);

	return TRUE;
}

void arvo_eventhandler_channel_stop(ArvoEventhandlerChannel *eventhandler_channel) {
	ArvoEventhandlerChannelPrivate *priv = eventhandler_channel->priv;
	if (priv->running) {
		g_source_remove(priv->event_source_id);
		priv->running = FALSE;
	}
}

static void finalize(GObject *object) {
	ArvoEventhandlerChannel *eventhandler_channel = ARVO_EVENTHANDLER_CHANNEL(object);
	arvo_eventhandler_channel_stop(eventhandler_channel);
	G_OBJECT_CLASS(arvo_eventhandler_channel_parent_class)->finalize(object);
}

static void arvo_eventhandler_channel_class_init(ArvoEventhandlerChannelClass *klass) {
	GObjectClass *gobject_class;

	gobject_class = (GObjectClass *)klass;

	gobject_class->finalize = finalize;

	g_type_class_add_private(klass, sizeof(ArvoEventhandlerChannelPrivate));
	
	signals[BUTTON] = g_signal_new("button", ARVO_EVENTHANDLER_CHANNEL_TYPE,
			G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
			0, NULL, NULL, g_cclosure_roccat_marshal_VOID__UCHAR_UCHAR, G_TYPE_NONE,
			2, G_TYPE_UCHAR, G_TYPE_UCHAR);
}
