/*
 * This file is part of roccat-tools.
 *
 * roccat-tools is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * roccat-tools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with roccat-tools. If not, see <http://www.gnu.org/licenses/>.
 */

#include "pyra_eventhandler_channel.h"
#include "g_cclosure_roccat_marshaller.h"
#include "pyra.h"
#include "roccat.h"
#include <sys/ioctl.h>

#define PYRA_EVENTHANDLER_CHANNEL_CLASS(klass) (G_TYPE_CHECK_CLASS_CAST((klass), PYRA_EVENTHANDLER_CHANNEL_TYPE, PyraEventhandlerChannelClass))
#define IS_PYRA_EVENTHANDLER_CHANNEL_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE((klass), PYRA_EVENTHANDLER_CHANNEL_TYPE))
#define PYRA_EVENTHANDLER_CHANNEL_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE((obj), PYRA_EVENTHANDLER_CHANNEL_TYPE, PyraEventhandlerChannelPrivate))

typedef struct _PyraEventhandlerChannelClass PyraEventhandlerChannelClass;
typedef struct _PyraEventhandlerChannelPrivate PyraEventhandlerChannelPrivate;

struct _PyraEventhandlerChannel {
	GObject parent;
	PyraEventhandlerChannelPrivate *priv;
};

struct _PyraEventhandlerChannelClass {
	GObjectClass parent_class;
};

struct _PyraEventhandlerChannelPrivate {
	guint event_source_id;
	gint report_size;
	gboolean running;
};

enum {
	CPI_CHANGED, /* cpi_index */
	PROFILE_CHANGED, /* profile_number */
	QUICKLAUNCH, /* key_index */
	MACRO,
	SHORTCUT,
	LAST_SIGNAL
};

static guint signals[LAST_SIGNAL] = { 0 };

G_DEFINE_TYPE(PyraEventhandlerChannel, pyra_eventhandler_channel, G_TYPE_OBJECT);

static void process_chrdev_event(PyraEventhandlerChannel *eventhandler_channel, PyraChrdevEvent const *event) {
	switch (event->type) {
	case PYRA_CHRDEV_EVENT_TYPE_PROFILE_2:
		g_signal_emit((gpointer)eventhandler_channel, signals[PROFILE_CHANGED], 0, event->value);
		break;
	case PYRA_CHRDEV_EVENT_TYPE_CPI:
		g_signal_emit((gpointer)eventhandler_channel, signals[CPI_CHANGED], 0, event->value);
		break;
	case PYRA_CHRDEV_EVENT_TYPE_MACRO:
		g_signal_emit((gpointer)eventhandler_channel, signals[MACRO], 0, event->key);
		break;
	case PYRA_CHRDEV_EVENT_TYPE_SHORTCUT:
		g_signal_emit((gpointer)eventhandler_channel, signals[SHORTCUT], 0, event->key);
		break;
	case PYRA_CHRDEV_EVENT_TYPE_QUICKLAUNCH:
		g_signal_emit((gpointer)eventhandler_channel, signals[QUICKLAUNCH], 0, event->key);
		break;
	case PYRA_CHRDEV_EVENT_TYPE_TILT:
	case PYRA_CHRDEV_EVENT_TYPE_MULTIMEDIA:
		/* just ignore these events */
		break;
	default:
		g_debug("unknown event = %02x%02x%02x", event->type, event->value, event->key);
	}
}

static gboolean io_cb(GIOChannel *channel, GIOCondition condition, gpointer user_data) {
	PyraEventhandlerChannel *eventhandler_channel = PYRA_EVENTHANDLER_CHANNEL(user_data);
	PyraEventhandlerChannelPrivate *priv = eventhandler_channel->priv;
	gchar *event;
	GIOStatus status;
	GError *error = NULL;
	gsize length;

	if (condition & G_IO_HUP)
		return FALSE;

	event = g_malloc(priv->report_size);

	status = g_io_channel_read_chars(channel, event, priv->report_size, &length, &error);

	if (status == G_IO_STATUS_AGAIN)
		goto ok;

	if (status == G_IO_STATUS_ERROR || status == G_IO_STATUS_EOF) {
		g_critical("error reading io_channel");
		goto error;
	}

	if (length != priv->report_size) {
		g_critical("wrong length %zd %s", length, error->message);
		goto error;
	}

	switch (priv->report_size) {
	case sizeof(PyraChrdevEvent):
		process_chrdev_event(eventhandler_channel, (PyraChrdevEvent const *)event);
		break;
	default:
		g_critical("wrong event size");
		goto error;
	}

ok:
	g_free(event);
	return TRUE; // keep on going
error:
	g_free(event);
	return FALSE;
}

PyraEventhandlerChannel *pyra_eventhandler_channel_new(void) {
	return PYRA_EVENTHANDLER_CHANNEL(g_object_new(PYRA_EVENTHANDLER_CHANNEL_TYPE, NULL));
}

static void pyra_eventhandler_channel_init(PyraEventhandlerChannel *channel) {
	PyraEventhandlerChannelPrivate *priv = PYRA_EVENTHANDLER_CHANNEL_GET_PRIVATE(channel);
	channel->priv = priv;
	priv->running = FALSE;
}

gboolean pyra_eventhandler_channel_start(PyraEventhandlerChannel *eventhandler_channel, gchar const *path, GError **error) {
	PyraEventhandlerChannelPrivate *priv = eventhandler_channel->priv;
	GIOChannel *channel;
	GIOStatus status;
	int fd;
	
	channel = g_io_channel_new_file(path, "r", error);
	if (!channel)
		return FALSE;

	status = g_io_channel_set_encoding(channel, NULL, error);
	if (status != G_IO_STATUS_NORMAL) {
		g_io_channel_unref(channel);
		return FALSE;
	}

	fd = g_io_channel_unix_get_fd(channel);
	if (ioctl(fd, ROCCATIOCGREPSIZE, &priv->report_size) == -1)
		priv->report_size = sizeof(PyraChrdevEvent);

	if (priv->report_size != sizeof(PyraChrdevEvent)) {
		g_set_error(error, G_FILE_ERROR, G_FILE_ERROR_FAILED, "ioctl reported unsupported event size %i", priv->report_size);
		g_io_channel_unref(channel);
		return FALSE;
	}

	priv->event_source_id = g_io_add_watch(channel, G_IO_IN | G_IO_HUP, (GIOFunc)io_cb, eventhandler_channel);
	/* g_io_add_watch increases reference count
	 * this passes ownership of the channel to the watch */
	g_io_channel_unref(channel);

	return TRUE;
}

void pyra_eventhandler_channel_stop(PyraEventhandlerChannel *eventhandler_channel) {
	PyraEventhandlerChannelPrivate *priv = eventhandler_channel->priv;
	if (priv->running) {
		g_source_remove(priv->event_source_id);
		priv->running = FALSE;
	}
}

static void finalize(GObject *object) {
	PyraEventhandlerChannel *eventhandler_channel = PYRA_EVENTHANDLER_CHANNEL(object);
	pyra_eventhandler_channel_stop(eventhandler_channel);
	G_OBJECT_CLASS(pyra_eventhandler_channel_parent_class)->finalize(object);
}

static void pyra_eventhandler_channel_class_init(PyraEventhandlerChannelClass *klass) {
	GObjectClass *gobject_class;

	gobject_class = (GObjectClass *)klass;

	gobject_class->finalize = finalize;

	g_type_class_add_private(klass, sizeof(PyraEventhandlerChannelPrivate));
	
	signals[CPI_CHANGED] = g_signal_new("cpi-changed", PYRA_EVENTHANDLER_CHANNEL_TYPE,
			G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
			0, NULL, NULL, g_cclosure_marshal_VOID__UCHAR, G_TYPE_NONE,
			1, G_TYPE_UCHAR);

	signals[PROFILE_CHANGED] = g_signal_new("profile-changed", PYRA_EVENTHANDLER_CHANNEL_TYPE,
			G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
			0, NULL, NULL, g_cclosure_marshal_VOID__UCHAR, G_TYPE_NONE,
			1, G_TYPE_UCHAR);

	signals[QUICKLAUNCH] = g_signal_new("quicklaunch", PYRA_EVENTHANDLER_CHANNEL_TYPE,
			G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
			0, NULL, NULL, g_cclosure_marshal_VOID__UCHAR, G_TYPE_NONE,
			1, G_TYPE_UCHAR);

	signals[MACRO] = g_signal_new("macro", PYRA_EVENTHANDLER_CHANNEL_TYPE,
			G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
			0, NULL, NULL, g_cclosure_marshal_VOID__UCHAR, G_TYPE_NONE,
			1, G_TYPE_UCHAR);

	signals[SHORTCUT] = g_signal_new("shortcut", PYRA_EVENTHANDLER_CHANNEL_TYPE,
			G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
			0, NULL, NULL, g_cclosure_marshal_VOID__UCHAR, G_TYPE_NONE,
			1, G_TYPE_UCHAR);
}
