/*
 * This file is part of roccat-tools.
 *
 * roccat-tools is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * roccat-tools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with roccat-tools. If not, see <http://www.gnu.org/licenses/>.
 */

#include "ryos_firmware_update.h"
#include "roccat_device_hidraw.h"
#include "ryos_device.h"
#include "ryos_control.h"
#include "g_roccat_helper.h"
#include <string.h>

static gboolean ryos_check_firmware_write(RoccatDevice *ryos, GError **error) {
	RyosControl *control;
	guint8 value;

	do {
		g_usleep(100 * G_ROCCAT_USEC_PER_MSEC);
		control = (RyosControl *)ryos_device_read(ryos, RYOS_COMMAND_FIRMWARE_WRITE_CONTROL, sizeof(RyosControl), error);
		if (!control)
			return FALSE;
		value = control->value;
		g_free(control);
		switch (value) {
		case RYOS_CONTROL_VALUE_STATUS_OK:
			return TRUE;
			break;
		case RYOS_CONTROL_VALUE_STATUS_BUSY:
			g_usleep(500 * G_ROCCAT_USEC_PER_MSEC);
			break;
		case RYOS_CONTROL_VALUE_STATUS_CRITICAL:
			g_set_error(error, G_FILE_ERROR, G_FILE_ERROR_INVAL, "Firmware write check got critical status");
			return FALSE;
			break;
		case RYOS_CONTROL_VALUE_STATUS_INVALID:
			g_set_error(error, G_FILE_ERROR, G_FILE_ERROR_INVAL, "Firmware write check got invalid status");
			return FALSE;
			break;
		default:
			g_set_error(error, G_FILE_ERROR, G_FILE_ERROR_FAILED, "Firmware write check unknown error");
			return FALSE;
			break;
		}
	} while(1);
}

static gboolean ryos_firmware_write(RoccatDevice *roccat_device, RyosFirmwareWrite const *buffer, GError **error) {
	gboolean retval;

	retval = roccat_device_hidraw_write(roccat_device, (gchar const *)buffer, sizeof(RyosFirmwareWrite), error);
	if (!retval)
		return FALSE;

	return ryos_check_firmware_write(roccat_device, error);
}

static guint16 calc_chunk_bytesum(guint8 const *data) {
	guint i;
	guint16 result = 0;

	for (i = 0; i < RYOS_FIRMWARE_CHUNK_SIZE; ++i)
		result += data[i];

	return result;
}

gboolean ryos_firmware_write_0(RoccatDevice *roccat_device, GError **error) {
	RyosFirmwareWrite data;

	data.command = RYOS_COMMAND_FIRMWARE_WRITE;
	data.type = RYOS_FIRMWARE_WRITE_TYPE_0;
	memset(&data.data, 0, RYOS_FIRMWARE_CHUNK_SIZE);
	data.data[2] = 0x08;
	data.data[3] = 0xAC;
	data.data[4] = 0x32;
	data.data[5] = 0x10;
	data.data[6] = 0x03;
	data.data[7] = 0xCB;
	data.data[9] = 0x01;

	return ryos_firmware_write(roccat_device, &data, error);
}

gboolean ryos_firmware_write_1(RoccatDevice *roccat_device, GError **error) {
	RyosFirmwareWrite data;

	data.command = RYOS_COMMAND_FIRMWARE_WRITE;
	data.type = RYOS_FIRMWARE_WRITE_TYPE_1;
	memset(&data.data, 0, RYOS_FIRMWARE_CHUNK_SIZE);
	data.data[2] = (guint8)RYOS_FIRMWARE_SIZE;
	data.data[3] = (guint8)(RYOS_FIRMWARE_SIZE >> 8);
	data.data[4] = (guint8)(RYOS_FIRMWARE_SIZE >> 16);
	data.data[5] = (guint8)(RYOS_FIRMWARE_SIZE >> 24);
	data.data[6] = 0x08;
	data.data[7] = 0xAC;
	data.data[8] = 0x22;
	data.data[9] = 0x88;

	return ryos_firmware_write(roccat_device, &data, error);
}

gboolean ryos_firmware_write_2(RoccatDevice *roccat_device, GError **error) {
	RyosFirmwareWrite data;

	data.command = RYOS_COMMAND_FIRMWARE_WRITE;
	data.type = RYOS_FIRMWARE_WRITE_TYPE_2;
	memset(&data.data, 0, RYOS_FIRMWARE_CHUNK_SIZE);
	data.data[2] = 0xEC;

	return ryos_firmware_write(roccat_device, &data, error);
}

gboolean ryos_firmware_write_3(RoccatDevice *roccat_device, guint16 number, guint8 const *chunk, GError **error) {
	RyosFirmwareWrite data;
	guint16 bytesum;

	bytesum = calc_chunk_bytesum(chunk);

	data.command = RYOS_COMMAND_FIRMWARE_WRITE;
	data.type = RYOS_FIRMWARE_WRITE_TYPE_3;
	memset(&data.data, 0, RYOS_FIRMWARE_CHUNK_SIZE);
	data.data[2] = (guint8)number;
	data.data[3] = (guint8)(number >> 8);
	data.data[4] = (guint8)bytesum;
	data.data[5] = (guint8)(bytesum >> 8);
	data.data[6] = 0x08;
	data.data[7] = 0xAC;
	data.data[8] = 0x22;
	data.data[9] = 0x88;

	return ryos_firmware_write(roccat_device, &data, error);
}

gboolean ryos_firmware_write_4(RoccatDevice *roccat_device, guint8 const *chunk, GError **error) {
	RyosFirmwareWrite data;

	data.command = RYOS_COMMAND_FIRMWARE_WRITE;
	data.type = RYOS_FIRMWARE_WRITE_TYPE_4;
	memcpy(&data.data, chunk, RYOS_FIRMWARE_CHUNK_SIZE);

	return ryos_firmware_write(roccat_device, &data, error);
}

gboolean ryos_firmware_write_f(RoccatDevice *roccat_device, GError **error) {
	RyosFirmwareWrite data;

	data.command = RYOS_COMMAND_FIRMWARE_WRITE;
	data.type = RYOS_FIRMWARE_WRITE_TYPE_F;
	memset(&data.data, 0, RYOS_FIRMWARE_CHUNK_SIZE);

	return ryos_firmware_write(roccat_device, &data, error);
}
